import json
from typing import Union, Type, cast, Any, TypeVar, Generic, TYPE_CHECKING, Tuple, Dict
from weakref import WeakValueDictionary


class OriginCached(type):
    def __init__(self, *args, **kwargs):
        self.__instance_map = WeakValueDictionary()
        super().__init__(*args, **kwargs)

    def __call__(self, origin):
        if origin in self.__instance_map:
            return self.__instance_map[origin]
        else:
            obj = super().__call__(origin)
            self.__instance_map[origin] = obj
            return obj


JsonSerializable = Union[int, float, bool, list, dict, str]
ItemType = TypeVar("ItemType", bound=JsonSerializable)


class JsonMeta(metaclass=OriginCached):
    __slots__ = ("__origin__", "__weakref__")

    def __init__(self, origin: Type[ItemType]):
        self.__origin__ = getattr(origin, "__origin__", origin)

    def __getitem__(self, t: Type[ItemType]) -> Type[ItemType]:
        return cast(Type[ItemType], JsonMeta(origin=t))

    def __type_check__(self, instance: Any) -> bool:
        return isinstance(instance, self.__origin__)

    def __typecast__(self, value: Any) -> Any:
        return json.loads(value)


Json = JsonMeta(origin=dict)


class SecretStr(str):
    def __repr__(self) -> str:
        return "SecretStr('**********')"

    def __str__(self) -> str:
        return repr(self)


T = TypeVar("T")


class DefaultTypeCast(Generic[T]):
    if TYPE_CHECKING:  # pragma: no cover
        # populated by the Generic, defined here to help IDEs only
        __orig_bases__: Tuple[Any, ...]

    @classmethod
    def __type_check__(cls, instance: Any) -> bool:
        return isinstance(instance, cls.__orig_bases__[0].__args__[0])

    @classmethod
    def __typecast__(cls, value: Any) -> T:
        raise TypeError(f"Object of type {cls.__name__} can not be typecast")


BOOL_STRINGS = {"true", "1", "yes", "y"}


class Boolean(DefaultTypeCast[bool]):
    @classmethod
    def __typecast__(self, value: Any) -> bool:
        if isinstance(value, str):
            return value.lower() in BOOL_STRINGS
        else:
            return bool(value)


DEFAULT_TYPE_CAST: Dict[Type, Type[DefaultTypeCast]] = {bool: Boolean}



import json
import logging
import threading
import time
from http import HTTPStatus
from typing import Dict

import requests

from ..configalchemy import BaseConfig, ConfigType

time_counter = time.time


class ConfigException(Exception):
    ...


class ApolloBaseConfig(BaseConfig):
    CONFIGALCHEMY_ENABLE_FUNCTION = True

    APOLLO_USING_CACHE = False
    APOLLO_SERVER_URL = ""
    APOLLO_APP_ID = ""
    APOLLO_CLUSTER = "default"
    APOLLO_NAMESPACE = "application"

    APOLLO_EXTRA_NAMESPACE = ""
    APOLLO_EXTRA_NAMESPACE_PRIORITY = 9

    APOLLO_LONG_POLL_TIMEOUT = 80

    def __init__(self):
        self.apollo_notification_map: Dict[str, ConfigType] = {}
        super().__init__()

    def start_long_poll(self):
        logging.info("start long poll")
        thread = threading.Thread(target=self.long_poll)
        thread.daemon = True
        thread.start()
        return thread

    def _access_config_by_namespace(self, namespace: str) -> ConfigType:
        route = "configs"
        if self.APOLLO_USING_CACHE:
            route = "configfiles"
        url = (
            f"{self.APOLLO_SERVER_URL}/{route}/{self.APOLLO_APP_ID}/"
            f"{self.APOLLO_CLUSTER}/{namespace}"
        )
        logging.info(f"Access apollo server url: {url}")
        response = requests.get(url)
        if response.ok:
            data = response.json()
            self.apollo_notification_map.setdefault(data["namespaceName"], {"id": -1})
            self.apollo_notification_map[data["namespaceName"]]["data"] = data.get(
                "configurations", {}
            )
            logging.debug(f"Got from apollo: {data}")
            return data.get("configurations", {})
        else:
            raise ConfigException(f"loading config failed: {url}")

    def configuration_function(self) -> ConfigType:
        self.from_mapping(
            self._access_config_by_namespace(self.APOLLO_NAMESPACE),
            priority=self.CONFIGALCHEMY_FUNCTION_VALUE_PRIORITY,
        )
        for namespace in self.APOLLO_EXTRA_NAMESPACE.split(","):
            if namespace:
                self.from_mapping(
                    self._access_config_by_namespace(namespace),
                    priority=self.APOLLO_EXTRA_NAMESPACE_PRIORITY,
                )
        return {}

    def long_poll_from_apollo(self):
        url = f"{self.APOLLO_SERVER_URL}/notifications/v2/"
        notifications = []
        for key, value in self.apollo_notification_map.items():
            notifications.append({"namespaceName": key, "notificationId": value["id"]})

        r = requests.get(
            url=url,
            params={
                "appId": self.APOLLO_APP_ID,
                "cluster": self.APOLLO_CLUSTER,
                "notifications": json.dumps(notifications, ensure_ascii=False),
            },
            timeout=self.APOLLO_LONG_POLL_TIMEOUT,
        )

        if r.status_code == HTTPStatus.NOT_MODIFIED:
            logging.info("Apollo No change, loop...")
        elif r.status_code == HTTPStatus.OK:
            data = r.json()
            for entry in data:
                logging.info(
                    "%s has changes: notificationId=%d"
                    % (entry["namespaceName"], entry["notificationId"])
                )
                namespace = entry["namespaceName"]
                if namespace == self.APOLLO_NAMESPACE:
                    self.from_mapping(
                        self._access_config_by_namespace(namespace),
                        priority=self.CONFIGALCHEMY_FUNCTION_VALUE_PRIORITY,
                    )
                else:
                    self.from_mapping(
                        self._access_config_by_namespace(namespace),
                        priority=self.APOLLO_EXTRA_NAMESPACE_PRIORITY,
                    )
                self.apollo_notification_map[entry["namespaceName"]]["id"] = entry[
                    "notificationId"
                ]
        else:  # pragma: no cover
            raise ConfigException(f"{url} : unexpected status {r.status_code}")

    def long_poll(self):
        while True:
            try:
                logging.debug("start apollo configuration long poll")
                self.long_poll_from_apollo()
            except ConfigException:
                time.sleep(5)



import io
import os.path
import sys
import traceback
from importlib import import_module
from typing import Any, Optional

currentframe = lambda: sys._getframe(2)

_srcfile = os.path.normcase(__file__)


def import_reference(reference: str) -> Any:
    module_path, class_name = reference.rsplit(".", 1)
    backend_cls = getattr(import_module(module_path), class_name)
    return backend_cls


def find_caller() -> Optional[str]:
    f = currentframe()
    stack_info = None
    while hasattr(f, "f_code"):
        co = f.f_code
        filename = os.path.normcase(co.co_filename)
        if filename == _srcfile:
            if f.f_back:
                f = f.f_back
                continue
        sio = io.StringIO()
        sio.write("Stack (most recent call last):\n")
        traceback.print_stack(f, file=sio)
        stack_info = sio.getvalue()
        if stack_info[-1] == "\n":
            stack_info = stack_info[:-1]
        sio.close()
        break
    return stack_info



import unittest

from configalchemy import BaseConfig
from configalchemy.utils import import_reference, find_caller


class UtilsTestCase(unittest.TestCase):
    def test_import_reference(self):
        self.assertEqual(
            id(BaseConfig),
            id(import_reference("configalchemy.configalchemy.BaseConfig")),
        )

    def test_find_caller(self):
        stack_info = find_caller()
        stack_str = "stack_info = find_caller()"
        self.assertEqual(stack_str, stack_info[-len(stack_str) :])


if __name__ == "__main__":
    unittest.main()



import os
from json import JSONEncoder
from typing import Any, List, MutableMapping

from configalchemy.field import Field
from configalchemy.utils import find_caller

ConfigType = MutableMapping[str, Any]

CONFIG_ALCHEMY_VERBOSITY = os.getenv("CONFIG_ALCHEMY_VERBOSITY", "")


class ConfigMetaItem:
    __slots__ = ("priority", "value", "setter")

    if CONFIG_ALCHEMY_VERBOSITY:

        def __init__(self, priority: int, value: Any):
            self.priority = priority
            self.value = value
            try:
                self.setter = find_caller()
            except ValueError:
                self.setter = None

    else:

        def __init__(self, priority: int, value: Any):
            self.priority = priority
            self.value = value
            self.setter = None

    def __repr__(self) -> str:
        return f"ConfigMetaItem(priority={self.priority}, value={self.value})"

    __str__ = __repr__


class ConfigMeta:
    __slots__ = ("field", "items")

    def __init__(self, default_value: Any, field: Field, priority: int = 0):
        self.field = field
        self.items: List[ConfigMetaItem] = [ConfigMetaItem(priority, default_value)]

    @property
    def value(self) -> Any:
        return self.items[-1].value

    def set(self, priority: int, value: Any) -> None:
        value = self.field.validate(value)
        length = len(self.items)
        item = ConfigMetaItem(priority, value)
        for index in range(length, 0, -1):
            if self.items[index - 1].priority <= priority:
                self.items.insert(index, item)
                break
        else:
            self.items.insert(0, ConfigMetaItem(priority, value))

    def __repr__(self) -> str:
        return repr(self.value)

    __str__ = __repr__

    def __enter__(self):
        """Prepare to implement optimistic raw lock if necessary"""
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        pass


class ConfigMetaJSONEncoder(JSONEncoder):
    def default(self, o) -> Any:
        if isinstance(o, ConfigMeta):
            return o.value
        return super().default(o)



import copy
import time
import unittest
from concurrent.futures.thread import ThreadPoolExecutor
from unittest.mock import MagicMock

from configalchemy.lazy import lazy, proxy, reset_lazy


class LazyTestCase(unittest.TestCase):
    def test_lazy_load(self):
        data = {}
        lazy_data = lazy(lambda: data)
        lazy_data["test"] = 1
        self.assertEqual(dir(data), dir(lazy_data))
        self.assertEqual(repr(data), repr(lazy_data))
        self.assertEqual(repr(data), repr(lazy_data))
        self.assertEqual(bool(data), bool(lazy_data))
        del lazy_data["test"]
        self.assertNotIn("test", lazy_data)
        with self.assertRaises(AttributeError):
            lazy_data.attr

    def test_lazy_load_load_once(self):
        call_mock = MagicMock()

        def get() -> int:
            time.sleep(0.1)
            call_mock()
            return 1

        number = lazy(get)

        def task(num):
            self.assertEqual(num + 1, number + num)

        with ThreadPoolExecutor(max_workers=4) as worker:
            for _ in worker.map(task, range(4)):
                _

        self.assertEqual(1, call_mock.call_count)
        object.__setattr__(number, "__attr__", None)

    def test_reset_lazy(self):
        call_mock = MagicMock()

        def get() -> int:
            call_mock()
            return 1

        number = lazy(get)

        self.assertEqual(1 + 1, number + 1)
        self.assertEqual(1, call_mock.call_count)
        reset_lazy(number)
        self.assertEqual(1 + 1, number + 1)
        self.assertEqual(2, call_mock.call_count)

    def test_lazy_load_operations_math(self):
        call_mock = MagicMock()

        def get() -> int:
            call_mock()
            return 2

        number = lazy(get)

        self.assertEqual(2, number)
        self.assertNotEqual(3, number)

        self.assertEqual(3, number + 1)
        self.assertEqual(3, 1 + number)

        self.assertEqual(-1, 1 - number)
        self.assertEqual(1, number - 1)

        self.assertEqual(2, number * 1)
        self.assertEqual(2, 1 * number)

        self.assertEqual(2, number / 1)
        self.assertEqual(0.5, 1.0 / number)

        self.assertEqual(2.0, number // 1)
        self.assertEqual(0.0, 1.0 // number)

        self.assertEqual(0, number % 2)
        self.assertEqual(0, 2 % number)

    def test_lazy_load_copy(self):
        class Foo:
            def __copy__(self):
                return self

            def __deepcopy__(self, memo):
                return self

        foo = Foo()
        lazy_foo = lazy(lambda: foo)
        foo.a = 1
        self.assertEqual(foo.__dict__, lazy_foo.__dict__)
        self.assertIs(foo, copy.copy(lazy_foo))
        self.assertIs(foo, copy.deepcopy(lazy_foo))

    def test_proxy_operations_math(self):
        call_mock = MagicMock()

        def get() -> int:
            call_mock()
            return 2

        number = proxy(get)

        self.assertEqual(2, number)
        self.assertNotEqual(3, number)

        self.assertEqual(3, number + 1)
        self.assertEqual(3, 1 + number)

        self.assertEqual(-1, 1 - number)
        self.assertEqual(1, number - 1)

        self.assertEqual(2, number * 1)
        self.assertEqual(2, 1 * number)

        self.assertEqual(2, number / 1)
        self.assertEqual(0.5, 1.0 / number)

        self.assertEqual(2.0, number // 1)
        self.assertEqual(0.0, 1.0 // number)

        self.assertEqual(0, number % 2)
        self.assertEqual(0, 2 % number)

        self.assertEqual(14, call_mock.call_count)


if __name__ == "__main__":
    unittest.main()



from typing import Any, Union, Callable

from configalchemy.types import DEFAULT_TYPE_CAST


class ValidateException(Exception):
    def __init__(self, name: str, value: Any):
        self.name = name
        self.value = value

    def __repr__(self) -> str:
        return f"""{self.name}'s value is invalid:

value: {self.value} 
type {type(self.value)}
"""

    def __str__(self) -> str:
        return repr(self)


class Field:
    __slots__ = (
        "name",
        "annotation",
        "value_type",
        "default_value",
        "type_check",
        "typecast",
    )

    def __init__(self, *, name: str, annotation: Any, default_value: Any):
        self.name = name
        self.annotation = annotation
        self.default_value = default_value
        self.value_type = type(default_value)
        self.type_check: Callable[[Any], bool] = self._type_check
        self.typecast: Callable[[Any], Any] = self._typecast

        self.prepare()

        if self.value_type in DEFAULT_TYPE_CAST:
            self.type_check = DEFAULT_TYPE_CAST[self.value_type].__type_check__
            self.typecast = DEFAULT_TYPE_CAST[self.value_type].__typecast__

        if getattr(self.default_value, "__type_check__", None):
            self.type_check = self.default_value.__type_check__
        if getattr(self.default_value, "__typecast__", None):
            self.typecast = self.default_value.__typecast__

        if getattr(self.annotation, "__type_check__", None):
            self.type_check = self.annotation.__type_check__
        if getattr(self.annotation, "__typecast__", None):
            self.typecast = self.annotation.__typecast__

    def prepare(self) -> None:
        origin = getattr(self.annotation, "__origin__", None)
        if origin is None:
            # field is not "typing" object eg. Union etc.
            return
        if origin is Union:
            self.value_type = self.annotation.__args__
            self.typecast = lambda x: self.value_type[0](x)
            return

    def validate(self, value: Any) -> Any:
        if self.type_check(value):
            return value
        else:
            try:
                return self.typecast(value)
            except Exception as e:
                raise ValidateException(self.name, value) from e

    def _type_check(self, value: Any) -> bool:
        return isinstance(value, self.value_type)

    def _typecast(self, value: Any) -> Any:
        return self.value_type(value)



import copy
from threading import Lock
from typing import TypeVar, Callable

LazyLoadType = TypeVar("LazyLoadType")


class LazyObject:
    __slots__ = ("__obj__", "__kwargs__", "__attr__", "__lock__", "__dict__")

    def __init__(self, obj: Callable[..., LazyLoadType], args, kwargs):
        object.__setattr__(self, "__obj__", obj)
        object.__setattr__(self, "__args__", args)
        object.__setattr__(self, "__kwargs__", kwargs)
        object.__setattr__(self, "__attr__", None)
        object.__setattr__(self, "__lock__", Lock())

    def __get_current_object__(self):
        # evaluated once on first access
        if self.__attr__ is None:
            with self.__lock__:
                if self.__attr__ is None:
                    object.__setattr__(
                        self,
                        "__attr__",
                        self.__obj__(*self.__args__, **self.__kwargs__),
                    )
                else:
                    return self.__attr__
        return self.__attr__

    def __getattr__(self, item):
        return getattr(self.__get_current_object__(), item)

    @property
    def __dict__(self):
        try:
            return self.__get_current_object__().__dict__
        except RuntimeError:
            raise AttributeError("__dict__")

    def __repr__(self):
        obj = self.__get_current_object__()
        return repr(obj)

    def __bool__(self):
        return bool(self.__get_current_object__())

    def __dir__(self):
        return dir(self.__get_current_object__())

    def __setitem__(self, key, value):
        self.__get_current_object__()[key] = value

    def __delitem__(self, key):
        del self.__get_current_object__()[key]

    __setattr__ = lambda x, n, v: setattr(
        x.__get_current_object__(), n, v  # type: ignore
    )
    __delattr__ = lambda x, n: delattr(x.__get_current_object__(), n)  # type: ignore
    __str__ = lambda x: str(x.__get_current_object__())  # type: ignore
    __lt__ = lambda x, o: x.__get_current_object__() < o
    __le__ = lambda x, o: x.__get_current_object__() <= o
    __eq__ = lambda x, o: x.__get_current_object__() == o  # type: ignore
    __ne__ = lambda x, o: x.__get_current_object__() != o  # type: ignore
    __gt__ = lambda x, o: x.__get_current_object__() > o
    __ge__ = lambda x, o: x.__get_current_object__() >= o
    __hash__ = lambda x: hash(x.__get_current_object__())  # type: ignore
    __call__ = lambda x, *a, **kw: x.__get_current_object__()(*a, **kw)
    __len__ = lambda x: len(x.__get_current_object__())
    __getitem__ = lambda x, i: x.__get_current_object__()[i]
    __iter__ = lambda x: iter(x.__get_current_object__())
    __contains__ = lambda x, i: i in x.__get_current_object__()
    __add__ = lambda x, o: x.__get_current_object__() + o
    __sub__ = lambda x, o: x.__get_current_object__() - o
    __mul__ = lambda x, o: x.__get_current_object__() * o
    __floordiv__ = lambda x, o: x.__get_current_object__() // o
    __mod__ = lambda x, o: x.__get_current_object__() % o
    __divmod__ = lambda x, o: x.__get_current_object__().__divmod__(o)
    __pow__ = lambda x, o: x.__get_current_object__() ** o
    __lshift__ = lambda x, o: x.__get_current_object__() << o
    __rshift__ = lambda x, o: x.__get_current_object__() >> o
    __and__ = lambda x, o: x.__get_current_object__() & o
    __xor__ = lambda x, o: x.__get_current_object__() ^ o
    __or__ = lambda x, o: x.__get_current_object__() | o
    __div__ = lambda x, o: x.__get_current_object__().__div__(o)
    __truediv__ = lambda x, o: x.__get_current_object__().__truediv__(o)
    __neg__ = lambda x: -(x.__get_current_object__())
    __pos__ = lambda x: +(x.__get_current_object__())
    __abs__ = lambda x: abs(x.__get_current_object__())
    __invert__ = lambda x: ~(x.__get_current_object__())
    __complex__ = lambda x: complex(x.__get_current_object__())
    __int__ = lambda x: int(x.__get_current_object__())
    __float__ = lambda x: float(x.__get_current_object__())
    __oct__ = lambda x: oct(x.__get_current_object__())
    __hex__ = lambda x: hex(x.__get_current_object__())
    __index__ = lambda x: x.__get_current_object__().__index__()
    __coerce__ = lambda x, o: x.__get_current_object__().__coerce__(x, o)
    __enter__ = lambda x: x.__get_current_object__().__enter__()
    __exit__ = lambda x, *a, **kw: x.__get_current_object__().__exit__(*a, **kw)
    __radd__ = lambda x, o: o + x.__get_current_object__()
    __rsub__ = lambda x, o: o - x.__get_current_object__()
    __rmul__ = lambda x, o: o * x.__get_current_object__()
    __rdiv__ = lambda x, o: o / x.__get_current_object__()
    __rtruediv__ = __rdiv__
    __rfloordiv__ = lambda x, o: o // x.__get_current_object__()
    __rmod__ = lambda x, o: o % x.__get_current_object__()
    __rdivmod__ = lambda x, o: x.__get_current_object__().__rdivmod__(o)
    __copy__ = lambda x: copy.copy(x.__get_current_object__())
    __deepcopy__ = lambda x, memo: copy.deepcopy(x.__get_current_object__(), memo)


class ProxyLazyObject(LazyObject):
    def __get_current_object__(self):
        # evaluated on every access
        return self.__obj__(*self.__args__, **self.__kwargs__)


def lazy(obj: Callable[..., LazyLoadType], *args, **kwargs) -> LazyLoadType:
    return LazyObject(obj, args, kwargs)  # type: ignore


def proxy(obj: Callable[..., LazyLoadType], *args, **kwargs) -> LazyLoadType:
    return ProxyLazyObject(obj, args, kwargs)  # type: ignore


def reset_lazy(obj):
    object.__setattr__(obj, "__attr__", None)
